#include "cal_svtdef.h"
#include "tk.h"
#include "error.h"
#include "reglist.h"
#include "cal_global.h"
#include "cal_functions.h"
#include "opcodes.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>



char* cal_fluff_record(char** begin_record, char** parameter_record, 
		       char** event_record) {

int nchips_top;
int dummy, nrun, nchips, xy_scan, compressed, n_expected, n_minimum;
int this_byte, prev_byte, ich;
int i, ishift, new_first, old_first, itemp, old_shift;
int bad, jshift, startbit, address, event_data, trcount, trtstamp;
int n_begin, n_parameter, n_event, n_new_record;
static char return_message[200];
int begin_chip[MAX_NUM_CHIP],end_chip[MAX_NUM_CHIP];
int old_begin_chip[MAX_NUM_CHIP],old_end_chip[MAX_NUM_CHIP];
int dir[MAX_NUM_CHIP], respond[MAX_NUM_CHIP], cr[MAX_NUM_CHIP];
int adr[MAX_NUM_CHIP];
int nsides, nbottom, first_respond_right, first_bottom, store_ishift;
int first_right;
char* temp;
/*----------------------------------------------------------------*/
/* get the number of bytes in the various records                 */
/*----------------------------------------------------------------*/
  dummy = cal_extract(1, &n_begin,      *begin_record);
  dummy = cal_extract(1, &n_parameter,  *parameter_record);
  dummy = cal_extract(1, &n_event,      *event_record);
/*----------------------------------------------------------------*/
/* check the number of bytes in the begin record                  */
/*----------------------------------------------------------------*/
  if (n_begin != (4+4+1+128)) {
    sprintf(return_message,"ERROR: Begin run record has %d bytes",n_begin);
    return return_message;
  }
/*----------------------------------------------------------------*/
/* extract nrun, number of chips                                  */
/*----------------------------------------------------------------*/
  dummy      = cal_extract(1, &nrun, (*begin_record+4));
  nchips     = (int) *(*begin_record+8);
  nchips     = nchips & 0xff;
  nchips_top = (int) *(*begin_record+10);
  nchips_top = nchips_top & 0xff;
/*----------------------------------------------------------------*/
/* see if this is an XY scan                                      */
/*----------------------------------------------------------------*/
  xy_scan = (int) *(*begin_record+9);
  xy_scan = xy_scan & 0x1;
/*----------------------------------------------------------------*/
/* make sure that this is a compressed run                        */
/*----------------------------------------------------------------*/
  compressed = (int) *(*begin_record+9);
  compressed = compressed & 0x2;
  if (compressed == 0) {
    sprintf(return_message,"ERROR: Attempting to fluff uncompressed data");
    return return_message;
  }
/*----------------------------------------------------------------*/
/* check the number of bytes in the parameter record              */
/*----------------------------------------------------------------*/
  n_expected = 4 + 1 + 36*nchips;
  if (n_parameter != n_expected) {
    sprintf(return_message,"ERROR: Parameter record has %d bytes, %d expected",
	    n_parameter,n_expected);
    return return_message;
  }
/*----------------------------------------------------------------*/
/* make sure the the compressed-event record is large enough      */
/*----------------------------------------------------------------*/
 n_minimum = 4 + 1 + 4 + nchips*4;
 if (xy_scan != 0) n_minimum = n_minimum + 8;
 if (n_event < n_minimum) {
    sprintf(return_message,
    "ERROR: Compressed run record has %d bytes, at least %d expected",
	    n_event,n_minimum);
    return return_message;
  }
/*----------------------------------------------------------------*/
/* find the pointers for the end chip data                        */
/*----------------------------------------------------------------*/
 ich = -1;
 ishift = 9;
 if (xy_scan != 0) ishift=ishift+8;
 for (i=ishift+3; i<n_event; i=i+2) {
   this_byte = (int) ( *(*event_record+i) );
   prev_byte = (int) ( *(*event_record+i-1) );
   this_byte = this_byte & 0xff;
   prev_byte = prev_byte & 0xff;
   if (prev_byte == 0x80) prev_byte = 0x0;   /* corrupted trailer */
   if (this_byte == 0 && prev_byte == 0) {
     ich++;
     if (ich > (nchips-1)) {
       sprintf(return_message,
	       "ERROR: found too many chips in compressed event record");
       return return_message;
     }
     end_chip[ich] = i;
   }
 }
/*----------------------------------------------------------------*/
/* Make sure that we found the right number of chips              */
/*----------------------------------------------------------------*/
 if (ich != (nchips-1)) {
   sprintf(return_message,
	       "ERROR: found %d chips in compressed event record..Expectd %d",
	       (ich+1),nchips);
   return return_message;
 }
/*----------------------------------------------------------------*/
/* Make sure that the end of the last chip is the end of data     */
/*----------------------------------------------------------------*/
 if (end_chip[ich] != (n_event-1)) {
   sprintf(return_message,
"ERROR: Found all chip data in compressed event record before end of record");
   return return_message;
 }
/*----------------------------------------------------------------*/
/* Get number of bytes in fluffed event record                    */
/*----------------------------------------------------------------*/
 n_new_record = 4 + 4 + 4 + 1 + 36*nchips + 8*nchips + (n_event-9);
 if (xy_scan != 0) n_new_record = n_new_record + 8;
/*----------------------------------------------------------------*/
/* copy compressed event to somewhere else                        */
/*----------------------------------------------------------------*/
 temp = (char *) malloc(n_event);
 if (temp == NULL) {
   sprintf(return_message,"ERROR: Allocating memory");
   return return_message;
 }
 for (i=0; i<n_event; i++) {
   *(temp+i) = *(*event_record+i);
/*   printf("byte %d = 0x%x \n",i,((int) (*(temp+i)) & 0xff));   */
 }
/*----------------------------------------------------------------*/
/* release memory for original compressed event record            */
/*----------------------------------------------------------------*/
 free(*event_record);
/*----------------------------------------------------------------*/
/* Allocate memory for uncompressed event record                  */
/*----------------------------------------------------------------*/
 *event_record = (char *) malloc(n_new_record);
 if (*event_record == NULL) {
   free(temp);
   sprintf(return_message,"ERROR: Allocating memory");
   return return_message;
 }
/*----------------------------------------------------------------*/
/* Start to fill the new record                                   */
/*----------------------------------------------------------------*/
 dummy = cal_extract(0, &n_new_record, *event_record);
 dummy = cal_extract(0, &nrun, (*event_record+4));
 for (i=0; i<4; i++) {                            /* event number  */
   *(*event_record+8+i) = *(temp+5+i);
 }
 *(*event_record+12) = (char) nchips;
 ishift = 13;
 for (i=0; i<(36*nchips); i++) {
   *(*event_record+ishift+i) = *(*parameter_record+5+i);
 }
 ishift = ishift + 36*nchips;  
/*----------------------------------------------------------------*/
/* get the beginning and end chip pointers in new record           */
/*----------------------------------------------------------------*/
 new_first = ishift + 8*nchips;
 old_first = 9;
 if (xy_scan != 0) {
   new_first = new_first + 16;
   old_first = old_first + 8;
 }
 for (i=0; i<nchips; i++) {
   end_chip[i] = new_first + end_chip[i] - old_first;
 }
 begin_chip[0] = new_first;
 for (i=1; i<nchips; i++) {
   begin_chip[i] = end_chip[i-1] + 1;
 }
/*----------------------------------------------------------------*/
/* put the begin and end chip pointers in                         */
/*----------------------------------------------------------------*/
 store_ishift = ishift;
 for (i=0; i<nchips; i++) {
   itemp = begin_chip[i];
   dummy  = cal_extract(0, &itemp, (*event_record+ishift));
   ishift = ishift + 4;
 }
 for (i=0; i<nchips; i++) {
   itemp = end_chip[i];
   dummy  = cal_extract(0, &itemp, (*event_record+ishift));
   ishift = ishift + 4;
 }
/*----------------------------------------------------------------*/
/* put the xy scan information pointers in                        */
/*----------------------------------------------------------------*/
 if (xy_scan != 0) {
   itemp  = ishift + 8;
   dummy  = cal_extract(0, &itemp, (*event_record+ishift));
   ishift = ishift + 4;
   itemp  = itemp + 4;
   dummy  = cal_extract(0, &itemp, (*event_record+ishift));
   ishift = ishift + 4;
 }
/*----------------------------------------------------------------*/
/* put the xy scan information in                                 */
/*----------------------------------------------------------------*/
 if (xy_scan != 0) {
   for (i=0; i<8; i++) {
     *(*event_record+ishift) = *(temp+9+i);
     ishift++;
   }
 }
/*----------------------------------------------------------------*/
/* FINALLY, put the chip data in                                  */
/*----------------------------------------------------------------*/
 old_shift = 9;
 if (xy_scan != 0) old_shift = old_shift + 8; 
 for (i=ishift; i<n_new_record; i++) {
   *(*event_record+i) = *(temp+old_shift);
   old_shift++;
 }
/*----------------------------------------------------------------*/
/* There was a bug in the compressed event record.  Basically     */
/* when reading from the right the order of the data is not       */
/* 0-1-2-3... as assumed here, but opposite.  The following       */
/* code hopefully fixes that.....                                 */
/* NB: The fix is not necessary after version 1.0.  We recognize  */
/* this case because the nchips_top variable will be filled       */
/*----------------------------------------------------------------*/
 if (nchips_top != 0) goto end;
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* first unpack control registers (direction and respond)  */
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
 ishift = 13;
 for (ich=0; ich<nchips; ich++) {
   dummy   = cal_extract(1, &itemp, (*event_record+ishift));
   cr[ich] = itemp;
   if ( (itemp & 0x1000000) == 0) {
     respond[ich] = 0;
   } else {
     respond[ich] = 1;
   }
   if ( (itemp & 0x800000) == 0) {
     dir[ich] = 0;
   } else {
     dir[ich] = 1;
   }
   ishift = ishift + 4;
} 
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* if we are always reading from the left, everything OK   */
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
   i = 0;
   for (ich=0; ich<nchips; ich++) {
     if (dir[ich] == 1) i=1;
   }
   if (i==0) goto end;
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* store the screwed up begin/end data pointers            */
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
   for (ich=0; ich<nchips; ich++) {
     old_begin_chip[ich] = begin_chip[ich];
     old_end_chip[ich]   = end_chip[ich];
   }
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* get the chip addresses.  These could be wrong           */
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
   for (ich=0; ich<nchips; ich++) {
     jshift = begin_chip[ich];
     cal_get_header(*event_record, jshift, &startbit, &address, 
		    &event_data, &trcount, &trtstamp);
     adr[ich] = address;
   }
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* get the number of sides                                 */
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
   nsides = 0;
   for (ich=0; ich<nchips; ich++) {
     if (adr[ich] == 0) nsides++;
   }
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* if it is a one sided HDI, easy to get number top/bottom */
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
   if (nsides == 0) {
     nchips_top    = nchips;
     nbottom       = 0;
     goto fix_top;
   }
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* It is a 2 sided HDI.  Look for the first chip which          */
/* reads out from the right and has respond to read set.        */
/* If this chip is not the very last chip, then this chip must  */
/* be the very last chip on the top of the HDI                  */
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
   first_respond_right = -1;
   for (ich=0; ich<nchips; ich++) {
     if (dir[ich] == 1 && respond[ich] == 1 && first_respond_right == -1)
       first_respond_right = ich;
   }
   if (first_respond_right != (nchips-1)) {
     nchips_top    = first_respond_right + 1;
     nbottom       = nchips - nchips_top;
     goto fix_top;
   }
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* If we are here, the top of the HDI must be set to read left. */
/* We can find the first chip on the bottom by scanning the     */
/* chips (skipping the 1st one) and looking for either          */
/*  - the first one with respond-to-read set  OR                */
/*  - the first one which reads to the right                    */ 
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
   first_bottom = -1;
   for (ich=1; ich<nchips; ich++) {
     if (dir[ich] == 1     && first_bottom == -1) first_bottom=ich;
     if (respond[ich] == 1 && first_bottom == -1) first_bottom=ich;
   }
   nchips_top    = first_bottom;
   nbottom       = nchips - nchips_top;
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* Okay, fixup the pointers
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
fix_top: {}
   first_right = nchips_top;
   for (ich=0; ich<nchips_top; ich++) {
     if (dir[ich] == 1 && first_right == nchips_top) first_right=ich;
   }
   i = nchips_top - 1;
   for (ich=first_right; ich<nchips_top; ich++) {
     begin_chip[ich] = old_begin_chip[i];
     end_chip[ich]   = old_end_chip[i];
     i--;
   }

   first_right = nchips;
   for (ich=nchips_top; ich<nchips; ich++) {
     if (dir[ich] == 1 && first_right == nchips) first_right=ich;
   }
   i = nchips - 1;
   for (ich=first_right; ich<nchips; ich++) {
     begin_chip[ich] = old_begin_chip[i];
     end_chip[ich]   = old_end_chip[i];
     i--;
   }
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* put the correct begin and end chip pointers in          */
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
 ishift = store_ishift;
 for (i=0; i<nchips; i++) {
   itemp = begin_chip[i];
   dummy  = cal_extract(0, &itemp, (*event_record+ishift));
   ishift = ishift + 4;
 }
 for (i=0; i<nchips; i++) {
   itemp = end_chip[i];
   dummy  = cal_extract(0, &itemp, (*event_record+ishift));
   ishift = ishift + 4;
 }
/*----------------------------------------------------------------*/
/* DONE.  Free memory, get out of here                            */
/* also fill the number of chips variables in cal_global.h        */
/*----------------------------------------------------------------*/
end: {}
   number_of_chips     = nchips;
   number_of_chips_top = nchips_top; 
   free(temp);
   return "OK";
}




