#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "error.h"
#include "reglist.h"
#include "cal_svtdef.h"
#include "cal_global.h"


int cal_dump(int, FILE*);
int read_record(FILE*, char**);
void display_help();
int decode_input(char*, int*);
char* cal_fluff_record(char**, char**, char**);
int cal_write_file(FILE**);
int cal_compress_event_record();

int main(int argc, char* argv[]) {

FILE* infile;
FILE* outfile;
int num, dummy, i, n, status, nev, numbeg;
int j, compressed, record_type;
char* begin_record;
char* begin_record2;
int flag = 0;
int nump;
int mev = 0;
char* tempp;
char* tempe;
char* parameter_record = NULL;
FILE* dump_file;
char* command_result;
char input[200];
char output_file[200];
int nfluff;
int wrote_begin = 0;
/*-----------------------------------------------------------*/
/* make sure that file has been specified                    */
/*-----------------------------------------------------------*/
  if (argc != 3) {
    printf("usage: fix_data_format_bug oldfile newfile \n");
    return 1;
  }
/*-----------------------------------------------------------*/
/* open file to be read                                      */
/*-----------------------------------------------------------*/
  infile = fopen(argv[1],"r");
  if (infile == NULL) {
    printf("cannot open file %s \n",argv[1]);
    return 1;
  }
/*-----------------------------------------------------------*/
/* look at the very first record                             */
/*-----------------------------------------------------------*/
  numbeg = read_record(infile, &cal_data_point);
  if (numbeg == -1) {
    printf("reached end of file \n");
    return 1;
  }
  if (numbeg < -1) {
    printf("error reading begin run record of file %s \n",argv[1]);
    return 1;
  }
/*-----------------------------------------------------------*/
/* if it is not a compressed file, there is no bug to fix    */
/* similarly, if the number_of_chips_top is there            */
/* there is no bug to fix                                    */
/*-----------------------------------------------------------*/
  compressed = (int) *(cal_data_point+9);
  compressed = compressed & 0x2;
  if (compressed == 0) {
    printf("file %s is not compress and does not have a bug \n",argv[1]);
    return 0;
  }
  dummy = (int) *(cal_data_point+10);
  dummy = dummy & 0xff;
  if (dummy != 0) {
    printf("file %s does not have a bug \n",argv[1]);
    return 0;
  }
/*-----------------------------------------------------------*/
/* open output file                                          */
/*-----------------------------------------------------------*/
  outfile = fopen(argv[2],"w");
  if (outfile == NULL) {
    printf("cannot open file %s \n",argv[2]);
    return 1;
  }
/*-----------------------------------------------------------*/
/* store the begin run record twice                          */
/*-----------------------------------------------------------*/
  begin_record  = (char *) malloc(numbeg);
  begin_record2 = (char *) malloc(numbeg);
  if (begin_record == NULL || begin_record2 == NULL) {
    printf("error allocating memory \n");
    return 1;
  }
  for (j=0; j<numbeg; j++) {
    *(begin_record+j)  = *(cal_data_point+j);
    *(begin_record2+j) = *(cal_data_point+j);
  }
/*-----------------------------------------------------------*/
/* free up the memory                                        */
/*-----------------------------------------------------------*/
  free(cal_data_point);
/*-----------------------------------------------------------*/
/* loop over records                                         */
/*-----------------------------------------------------------*/
  for (;;) {
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - -*/
/* read the next record                                       */
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - -*/
    num = read_record(infile, &cal_data_point);
    if (num == -1) {
      printf("reached end of file %s \n", argv[1]);
      goto end;
    }
    if (num < -1) {
      printf("error reading record from file %s \n", argv[1]);
      return 1;
    }
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - -*/
/* see what kind of record it is                              */
/* 1 = parameter record                                       */
/* 2 = compressed event record                                */
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - -*/
    dummy = (int) *(cal_data_point+4);
    dummy = dummy & 0xff;

    if (dummy == 0x1) {
      nump = num;
      record_type = 1;
      if (parameter_record != NULL) free(parameter_record);
      parameter_record = (char *) malloc(nump);
      if (parameter_record == NULL) {
	printf("error allocating memory \n");
	return 1;
      }
      for (j=0; j<nump; j++) *(parameter_record+j) = *(cal_data_point+j);
      
    } else if (dummy == 0x2) {
      record_type = 2;
      if (parameter_record == NULL) {
	printf("error: compressed event record before any parameter record\n");
	return 1;
      }
      mev++;
      if ( (mev%1000) == 0) printf("processing event %d \n", mev);
    } else {
      printf("error: unknown record type \n");
      return 1;
    }

/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - -*/
/* if this is an event record in a compressed file, fluff it up   */
/* NB this will also fill the variables number_of_chips and       */
/*    number_of_chips_top in cal_global.h and fix the bug         */
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - -*/
    if (record_type == 2) {
      command_result = 
	cal_fluff_record(&begin_record, 
			 &parameter_record, 
			 &cal_data_point);
      if (strcmp("OK",command_result) != 0) {
	printf("%s \n",command_result);
	return 1;
      }
      dummy = cal_extract(1, &nfluff, cal_data_point);
    }

/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
/* we will write records as they come.  but we want to modify the   */
/* begin event record which we write out to include the nchips_top  */
/* variable.  this will flag the file as being bug free.  but we    */
/* cannot do this until we have the nchips_top variable from the    */
/* first compressed event record ....... what a mess .....          */
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/

/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
/* if it is a parameter record and the 1st two records have already */
/* been written out, just write it out                              */
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
    if (record_type == 1 && wrote_begin == 1) {
      dummy = cal_write_file(&outfile);
      if (dummy != 0) {
	printf("error writing to %s \n",argv[2]);
	return 1;
      }
      free(cal_data_point);
      goto next_record;
    }

/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
/* if it is a compressed event record and the 1st two records       */
/* have yet to be written out, do it                                */
/* (1)  store the fluffed event record in "tempe"                   */
/* (2)  store number_of_chips_top in "begin_record2"                */
/* (3)  copy "begin_record2" into "cal_data_point"                  */
/* (4)  write out the begin event record                            */
/* (5)  copy the parameter record into "cal_data_point"             */ 
/* (6)  write out the parameter record                              */
/* (7)  copy back "tempe" into "cal_data_point"                     */
/* (8)  get read of "tempe" and "begin_record2"                     */ 
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
    if (record_type == 2 && wrote_begin == 0) {
      wrote_begin = 1;

/*(1)*/
      tempe = (char *) malloc(nfluff);
      if (tempe == NULL) {
	printf("run out of memory \n");
	return 1;
      }
      for (j=0; j<nfluff; j++) *(tempe+j) = *(cal_data_point+j);
    
/*(2)*/
      *(begin_record2 + 10) = (char) number_of_chips_top;

/*(3)*/
      free(cal_data_point);
      cal_data_point = (char *) malloc(numbeg);
      if (cal_data_point == NULL) {
	printf("run out of memory \n");
	return 1;
      }
      for (j=0; j<numbeg; j++) *(cal_data_point+j) = *(begin_record2+j);

/*(4)*/
      dummy = cal_write_file(&outfile);
      if (dummy != 0) {
	printf("error writing to %s \n",argv[2]);
	return 1;
      }

/*(5)*/
      free(cal_data_point);
      cal_data_point = (char *) malloc(nump);
      if (cal_data_point == NULL) {
	printf("run out of memory \n");
	return 1;
      }
      for (j=0; j<nump; j++) *(cal_data_point+j) = *(parameter_record+j);

/*(6)*/
      dummy = cal_write_file(&outfile);
      if (dummy != 0) {
	printf("error writing to %s \n",argv[2]);
	return 1;
      }

/*(7)*/
      free(cal_data_point);
      cal_data_point = (char *) malloc(nfluff);
      if (cal_data_point == NULL) {
	printf("run out of memory \n");
	return 1;
      }
      for (j=0; j<nfluff; j++) *(cal_data_point+j) = *(tempe+j);

/*(8)*/
      free(tempe);
      free(begin_record2);
    }
      

/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
/* okay, now if it is event record, compress it, then write it out  */    
/* then release the memory that we have used                        */
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
    if (record_type == 2) {
      dummy = cal_compress_event_record();
      if (dummy != 0) {
	printf("error compressing the event record \n");
	return 1;
      }
      dummy = cal_write_file(&outfile);
      if (dummy != 0) {
	printf("error writing to %s \n",argv[2]);
	return 1;
      }
      free(cal_data_point);
    }
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
/* next record                                                      */
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
next_record: {}
  }
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
/* done                                                             */
/*- - - - - - - - - - - - - - - - - - -- - - - - - - - - - - - - - -*/
end: {}
  return 0;
}
