/**********************************************************/
/* Read event routine                                     */
/*   flag=1 open file, read header, ask for               */
/*          format checks customization                   */
/*   flag=10 open file, read header, do not ask for       */
/*          format checks customization                   */
/*   flag=2 read event                                    */
/* return  0 if OKAY                                      */
/* return -1 if we have succesfully reached end-of-file   */
/* Error codes :  1  called with illegal flag             */
/*                2  cannot open the file                 */
/*                3  error allocating memory              */
/*                4  error reading header size            */
/*                5  error unpacking header size          */
/*                6  error reading header                 */
/*                7  error converting array C-->FORTRAN   */
/*                8  error reading event record size      */
/*                9  error unpacking event record size    */
/*               10  error reading event record           */
/*               11  bad run number in event record       */
/*               12  bad number of chips in event record  */
/*               13  error reported from format_check     */
/*               14  bad record type                      */
/*               15  error uncompressing record           */
/**********************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "cal_svtdef.h"
#include "cal_format_check.h"

#define MAXHIT 128*MAX_NUM_CHIP

int ignore[MAX_NUM_CHIP];         /* needed by cal_format_check */

int  array_fixup(int*, int,  int);
void cal_get_header(char*, int, int*, int*, int*, int*, int*);
void cal_get_hit(char*, int, int*, int*, int*);
char* cal_format_check(char*, int);
int  cal_extract(int,  int*, char*);
int  switch_bits(int, int);
int read_record(FILE*, char**);
char* cal_fluff_record(char**, char**, char**);
int ask_user_util(char*);



int read_event_(flag, input_file, s1, s2, n1) 
int* flag;
char* input_file;
int n1;
struct svt_global_common {
  int nchip;
  int nrun;
  int nev;
  int calib_mask[8][MAX_NUM_CHIP];
  int channel_mask[8][MAX_NUM_CHIP];
  int icreg[MAX_NUM_CHIP];
} *s1;
struct svt_hit_common {
  int nhit;
  int ichipnum[MAXHIT];
  int ichan[MAXHIT];
  int itot[MAXHIT];
  int itstamp[MAXHIT];
  int itrcount[MAXHIT];
  int itrtstamp[MAXHIT];
  int ical_dac[MAXHIT];
  int ithr_dac[MAXHIT];
  int ical_msk[MAXHIT];
  int icha_msk[MAXHIT];
} *s2;

{
  static int check_flag; 
  static FILE *fp;
  char* temp;
  char* temp2;
  int nread;
  int num, ichip, mask, itemp;
  int i, j, k, ishift, value, dummy, ii, jbit, status;
  int begin_data[MAX_NUM_CHIP];
  int end_data[MAX_NUM_CHIP];
  int startbit, address, event_data, trcount, trtstamp;
  int nhits_this_chip;
  static int calib_mask_store[8][MAX_NUM_CHIP];
  static int channel_mask_store[8][MAX_NUM_CHIP];
  char* command_result;
  int record_type;
  static int compressed;
  static char* begin_record;
  static char* parameter_record;
/*-------------------------------------------------------------*/
/* Check for a valid flag                                      */
/*-------------------------------------------------------------*/
    if (*flag != 1 && *flag != 2 && *flag != 10) return 1;
/*-------------------------------------------------------------*/
/* Open the file                                               */
/* Append a null character to the file name                    */
/* Also, ask for what checks we want to do                     */
/*-------------------------------------------------------------*/
    if (*flag == 1 || *flag == 10) {
      check_flag = CHECK_ALL;
      if (*flag == 1) check_flag = ask_user();
      for (i=0; i<n1; i++) {
	if (input_file[i] == ' ') input_file[i] = NULL;
      } 
      fp = fopen(input_file,"r");
      if (fp == NULL) return 2;
    }
/*-------------------------------------------------------------*/
/* read the record, stick it in temp2                          */
/*-------------------------------------------------------------*/
 read_again:
    num = read_record(fp, &temp2);
    if (*flag == 1  && num <   0) return 6;
    if (*flag == 10 && num <   0) return 6;
    if (*flag == 2  && num <  -1) return 10;
    if (*flag == 2  && num == -1) return -1;
/*-------------------------------------------------------------*/
/* now unpack the begin run record                             */
/* code stolen from oasis_new/src/cal/cal_dump.c               */
/*-------------------------------------------------------------*/
  if (*flag == 1 || *flag == 10) {
    dummy  = cal_extract(1,&(s1->nrun),(temp2+4));
    s1->nchip  = (int) *(temp2+8);
  }
/*-------------------------------------------------------------*/
/* for begin run record, see if data file is compressed        */
/* if so, save the begin run record for later usage            */
/*-------------------------------------------------------------*/
  if (*flag == 1 || *flag == 10) {
    compressed = (int) *(temp2+9);
    compressed = compressed & 0x2;
    if (compressed != 0) {
      begin_record = (char *) malloc(num);
      if (begin_record == NULL) return 3;
      for (i=0; i<num; i++) {
	*(begin_record+i) = *(temp2+i);
      }
    }
  }	
/*-------------------------------------------------------------*/
/* fill the array of ignored chips on begin run record         */
/*-------------------------------------------------------------*/
  if (*flag == 1 || *flag == 10) {
    for (ichip=0; ichip<MAX_NUM_CHIP;  ichip++) {

      dummy = ichip / 8;
      i     = 11 + dummy;
      mask  = *(temp2+11+i) & 0xff;
      itemp = 1;
      if ( (ichip-8*dummy) != 0) itemp = itemp << (ichip-8*dummy);

      if ( (mask&itemp) != 0) {
	ignore[ichip] = 1;
      } else {
	ignore[ichip] = 0;
      }

    }
  }
/*-------------------------------------------------------------*/
/* If we were asked to read the begin run record, we are done  */
/*-------------------------------------------------------------*/
  if (*flag == 1 || *flag == 10) {
    parameter_record = NULL;
    free(temp2);
    return 0;
  }
/*----------------------------------------------------------------*/
/* if this is NOT a begin run record, and data file is compressed */
/* get the record type.                                           */
/*                       2=normal event                           */
/*                       3=compressed event                       */
/*                       4=parameter record                       */
/*----------------------------------------------------------------*/
  record_type = 2;
  if (compressed != 0) {
    j = (int) * (temp2+4);
    j = j & 0xff;
    if (j == 0x1) {
      record_type = 4;
    } else if (j == 0x2) {
      record_type = 3;
    } else {                     /* bad record type */
      free(temp2);
      return 14;
    }
  }
/*-----------------------------------------------------------------*/
/* if this is a parameter record:                                  */
/* (1) get rid of the previous parameter record                    */
/* (2) store the current parameter record                          */
/* (3) free the memory                                             */
/* (4) go back and read the next record                            */
/*-----------------------------------------------------------------*/
  if (record_type == 4) {
    if (parameter_record != NULL) free(parameter_record);
    parameter_record = (char *) malloc(num);
    if (parameter_record == NULL) {
      free(temp2);
      return 3;
    }
   for (i=0; i<num; i++) {
     *(parameter_record+i) = *(temp2+i);
   }
   free(temp2);
   goto read_again;
  }
/*-----------------------------------------------------------------*/
/* if this is a compressed event record, fluff it up               */
/*-----------------------------------------------------------------*/
  if (record_type == 3) {
    command_result = cal_fluff_record(&begin_record,&parameter_record,&temp2);
    if (strcmp("OK",command_result) != 0) {
      free(temp2);
      return 15;
    }
  }
/*-------------------------------------------------------------*/
/* check the format                                            */
/*-------------------------------------------------------------*/
  command_result = cal_format_check(temp2, check_flag);
  if (strcmp("OK",command_result) != 0) {
    printf("%s \n",command_result);
    return 13;
  }
/*-------------------------------------------------------------*/
/* run number, event number, number of chips                   */
/*-------------------------------------------------------------*/
  dummy = cal_extract(1, &value, (temp2+4));
  if (value != s1->nrun) return 11;
  dummy = cal_extract(1, &(s1->nev),  (temp2+8));
  value = (int) *(temp2+12);
  if (value != s1->nchip) return 12;
/*-------------------------------------------------------------*/
/* control registers                                           */
/*-------------------------------------------------------------*/
  ishift = 13;
  for (i=0; i<s1->nchip; i++) {
         dummy = cal_extract(1, &value,  (temp2+ishift));
         s1->icreg[i] = value;
	 ishift = ishift + 4;
  }
/*-------------------------------------------------------------*/
/* masks                                                       */
/*-------------------------------------------------------------*/
  for (i=0; i < s1->nchip; i++) {
       for (j=0; j<4; j++) {
          dummy = cal_extract(1,&value,(temp2+ishift));
	  s1->calib_mask[2*j][i]   = 0xffff & value;
	  s1->calib_mask[2*j+1][i] = (value >> 16) & 0xffff;
          ishift = ishift + 4;
        }
  }
  for (i=0; i < s1->nchip; i++) {
       for (j=0; j<4; j++) {
          dummy = cal_extract(1,&value,(temp2+ishift));
	  s1->channel_mask[2*j][i]   = 0xffff & value;
	  s1->channel_mask[2*j+1][i] = (value >> 16) & 0xffff;
          ishift = ishift + 4;
        }
  }
/*-------------------------------------------------------------*/
/* store the channel_mask and calib_mask in temp array         */
/*-------------------------------------------------------------*/
  for (i=0; i<s1->nchip; i++) {
      for (j=0; j<8; j++) {
	  calib_mask_store[j][i]   = s1->calib_mask[j][i];
	  channel_mask_store[j][i] = s1->channel_mask[j][i];
      }
  }
/*-------------------------------------------------------------*/
/* Fixup arrays so that they are seen correctly by FORTRAN     */
/* Fixed a bug    August 5 1998                                */
/*-------------------------------------------------------------*/
  /*  if (*flag == 1) {                bug Aug 5 1998      */
     dummy = array_fixup(*(s1->calib_mask),8,MAX_NUM_CHIP);
     if (dummy != 0) return 7;
     dummy = array_fixup(*(s1->channel_mask),8,MAX_NUM_CHIP);
     if (dummy != 0) return 7;
  /*  }                                bug Aug 5 1998      */
/*-------------------------------------------------------------*/
/* Get begin and end of data for each chip                     */
/*-------------------------------------------------------------*/
   for (i=0; i<s1->nchip; i++) {
         dummy = cal_extract(1, &value, (temp2+ishift));
         begin_data[i] = value;
         ishift = ishift + 4;
   }
   for (i=0; i<s1->nchip; i++) {
         dummy = cal_extract(1, &value, (temp2+ishift));
         end_data[i] = value;
         ishift = ishift + 4;
   }
/*-------------------------------------------------------------*/
/* Now loop over chips, NB if only four bytes ==> no hits      */
/* First two bytes are header                                  */
/* Last two bytes are trailer                                  */
/*-------------------------------------------------------------*/
   s2->nhit = 0;
   for (i=0; i<s1->nchip; i++) {
       nhits_this_chip = (end_data[i] - begin_data[i] - 3) / 2;
       ishift = begin_data[i];
       cal_get_header(temp2, ishift, &startbit, &address, &event_data, 
		             &trcount, &trtstamp);
       for (j=1; j<=nhits_this_chip; j++) {
	 s2->nhit = s2->nhit + 1;
	 k = s2->nhit - 1;
	 ishift = begin_data[i] + 2*j;
	 cal_get_hit(temp2, ishift, &(s2->ichan[k]), &(s2->itot[k]), 
		             &(s2->itstamp[k]));
	 s2->ichipnum[k]  = i;
	 s2->itrcount[k]  = trcount;
	 s2->itrtstamp[k] = trtstamp;
	 s2->ithr_dac[k]  = s1->icreg[i] & 0x3f;
	 s2->ical_dac[k]  = (s1->icreg[i] >> 6) & 0x3f;
	 ii    = s2->ichan[k] / 16;
	 jbit  = s2->ichan[k] - 16*ii;
         value = calib_mask_store[ii][i];
	 if (jbit != 0) value=value>>jbit;
	 s2->ical_msk[k]  = value & 0x1;
         value = channel_mask_store[ii][i];
	 if (jbit != 0) value=value>>jbit;
	 s2->icha_msk[k]  = value & 0x1;
       }
    }
/*-------------------------------------------------------------*/
/* Done, return                                                */
/*-------------------------------------------------------------*/
  free(temp2);
  return 0;

}

/*******************************************************************/
/* Array fixup.  In FORTRAN the first dimension of an array varies */
/* fastest, in C the last dimension varies the fastest             */
/* Take 2D int array in C, turn it into FORTRAN                    */
/* return 0 if OK                                                  */
/*******************************************************************/
int array_fixup(int* array, int nx, int ny) {

int* temp;
int i;
int j;
int c_shift;
int fortran_shift;

/*-------------------------------------------------------*/
/* Get temporary storage for array                       */
/*-------------------------------------------------------*/
   temp = (int *) malloc(nx*ny*sizeof(int));
   if (temp == NULL) return 1;
/*-------------------------------------------------------*/
/* store it the way that FORTAN likes it                 */
/*-------------------------------------------------------*/
   for (i=0; i<nx; i++) {
     for (j=0; j<ny; j++) {
       c_shift       = j + ny*i;
       fortran_shift = i + nx*j;
       *(temp+fortran_shift) = *(array+c_shift);
     }
   }
/*-------------------------------------------------------*/
/* overwite the old array                                */
/*-------------------------------------------------------*/
   for (i=0; i<nx*ny; i++) {
     *(array+i) = *(temp+i);
   }
/*-------------------------------------------------------*/
/* free the memory                                       */
/*-------------------------------------------------------*/
  free(temp);
/*-------------------------------------------------------*/
/* return                                                */
/*-------------------------------------------------------*/
  return 0;
}

/*-------------------------------------------------------*/
/* ask the user what checks she wants to make            */
/*-------------------------------------------------------*/
int ask_user() {
  char line[200];
  int flag;

  flag = CHECK_ALL;
  printf("\n As a default the code does a bunch of format checks \n");
  printf("Do you want to customize these checks (y/n)?\n");
  scanf("%s",line);
  if ( line[0] == 'y' || line[0] == 'Y') {
     flag = 0;
     if (ask_user_util("start bit") == 1) 
                   flag = flag | START_BIT;
     if (ask_user_util("for all chips reporting") == 1)
                   flag = flag | ALL_CHIPS_REPORTING;
     if (ask_user_util("data status bit") == 1)
                   flag = flag | EVENT_DATA_STATUS;
     if (ask_user_util("trigger tag") == 1)
                   flag = flag | TRIGGER_TAG;
     if (ask_user_util("trigger time") == 1)
                   flag = flag | TRIGGER_TIME;
     if (ask_user_util("for valid channel number") == 1)
                   flag = flag | VALID_CHANNEL_NUMBER;
     if (ask_user_util("for no repeat channel") == 1)
                   flag = flag | NO_REPEAT_CHANNEL;
     if (ask_user_util("against zero TOT") == 1)
                   flag = flag | NON_ZERO_TOT;
     if (ask_user_util("trailer") == 1)
                   flag = flag | TRAILER;
     if (ask_user_util("for even number of bytes") == 1)
                   flag = flag | NUMBER_OF_BYTES;
     if (ask_user_util("for valid chip addresses") == 1)
                   flag = flag | CHIP_ADDRESS;
     printf("Bit mask of enabled checks = 0x%x \n\n",flag);
   } 
   return flag;
}
int ask_user_util(char* question) {
  char line[200];
  int code;
  printf("Do you want to check %s ? (y/n) \n",question);
  scanf("%s",line);
  if ( line[0] == 'y' || line[0] == 'Y') {
     code = 1;
   } else {
     code = 0;
   }
   return code;
}

